from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from datetime import datetime
from itertools import izip_longest, imap
import os
from os.path import splitext
import sys


__all__ = [
    'dump',
    'grouper',
    'initalise_dirac',
]


def initalise_dirac():
    for module in ['XRootD', 'LHCbDIRAC', 'DIRAC']:
        if module in sys.modules:
            print('WARNING: urban_baracle should be imported before', module)
            print('Access to DIRAC may not function correctly')

    try:
        from LHCbDIRAC.DataManagementSystem.Client.DMScript import Script  # NOQA
    except ImportError:
        raise ImportError('Unable to import LHCbDIRAC, try using "lb-run LHCbDIRAC/prod python -m manci ..."')

    argv, sys.argv = sys.argv, []
    Script.parseCommandLine()
    os.environ['XrdSecPROTOCOL'] = 'gsi,unix,krb5'
    sys.argv = argv
    # TODO Check for grid proxy


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3) --> ABC DEF G
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return imap(lambda x: list(filter(bool, x)),
                izip_longest(fillvalue=fillvalue, *args))


def dump(obj, filename):
    """Dump an object to a file

    Dumps obj to filename, with automated detection for the filetype and
    encoding of datetime objects.
    """
    _, ext = splitext(filename)

    encoded_obj = {
        lfn: {k: (v.isoformat() if isinstance(v, datetime) else v)
              for k, v in metadata.items()}
        for lfn, metadata in obj.items()
    }

    if ext == '.json':
        import json
        with open(filename, 'wt') as fp:
            json.dump(encoded_obj, fp, default=make_object_serialisable)
    elif ext == '.yaml':
        import yaml
        with open(filename, 'wt') as fp:
            yaml.dump(encoded_obj, fp, default=make_object_serialisable)
    else:
        raise NotImplementedError('Unrecognised file extension ' + ext)


def make_object_serialisable(obj):
    from .structures import Replica

    if isinstance(obj, Replica):
        return {
            'SE': obj._se,
            'PFN': obj._pfn,
        }
    else:
        raise NotImplementedError(obj.__class__)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import json
from os.path import join
import re

from .data_management import lookup_files_from_query
from .data_management import loookup_replicas
from .data_management import mirror_files
from .utils import dump


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    mirror_parser = subparsers.add_parser('lookup')
    mirror_parser.add_argument('--mirror-dir', required=False)
    mirror_parser.add_argument('--output-fn', default=None)
    group = mirror_parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--production-id', default=None)
    group.add_argument('--bk-path', default=None)
    group.add_argument('--lfns', nargs='+', default=None)
    group.add_argument('--lfns-fn', default=None)
    group.add_argument('--batch', default=None)
    mirror_parser.set_defaults(func=lambda a: lookup(
        mirror_dir=a.mirror_dir, output_fn=a.output_fn, bk_path=a.bk_path, lfns=a.lfns,
        lfns_fn=a.lfns_fn, production_id=a.production_id, batch=a.batch))

    args = parser.parse_args()
    args.func(args)


def lookup(mirror_dir=None, output_fn=None, batch=None,
           bk_path=None, lfns=None, lfns_fn=None, production_id=None):
    if sum(map(bool, [batch, bk_path, lfns, lfns_fn, production_id])) != 1:
        raise ValueError('One and only one of "bk_path", "lfns", "lfns_fn", '
                         '"production_id" and "batch" should be passed')

    if batch is not None:
        if output_fn:
            raise NotImplementedError('output_fn cannot be used with batch')

        options = {
            'BKQUERY': 'bk_path',
            'LFN': 'lfns',
            'LFNs': 'lfns_fn',
            'PRODID': 'production_id',
        }

        with open(batch, 'rt') as fp:
            jobs = {k: v if isinstance(v, list) else [v]
                    for k, v in json.load(fp).items()}

        for directory, commands in jobs.items():
            for command in commands:
                directory = join(mirror_dir, directory)
                match = re.match(r'^(BKQUERY|LFN|LFNs|PRODID):(.*)$', command)
                if match is None:
                    raise ValueError('Invalid command "' + command + '"')
                lookup(directory, **{options[match.group(1)]: match.group(2)})

        return

    if bk_path is not None:
        if not isinstance(bk_path, list):
            bk_path = [bk_path]
        files = lookup_files_from_query(bk_path)

    if lfns_fn is not None:
        with open(lfns_fn, 'rt') as fp:
            files = {lfn.strip(): {} for lfn in fp.read().split('\n')}

    if lfns is not None:
        if not isinstance(lfns, list):
            lfns = [lfns]
        files = {lfn: {} for lfn in lfns}

    if production_id is not None:
        if not isinstance(production_id, list):
            production_id = [production_id]
        raise NotImplementedError()

    loookup_replicas(files)
    if output_fn:
        dump(files, output_fn)

    if mirror_dir:
        mirror_files(files, mirror_dir)


if __name__ == '__main__':
    parse_args()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from os.path import join

from XRootD import client
from tqdm import tqdm

from .exceptions import ChecksumMismatchError


__all__ = [
    'URL',
    'CopyProcess',
]


class URL(object):
    _filesystems = {}

    def __init__(self, path):
        self._path = path
        self._url = client.URL(path)
        assert self._url.is_valid(), path

    def __str__(self):
        return self._path

    def __repr__(self):
        return '{name}({path})'.format(name=self.__class__, path=self._path)

    @property
    def _fs(self):
        host = self._url.protocol + '://' + self._url.hostid
        if host not in self._filesystems:
            self._filesystems[host] = client.FileSystem(host)
        return self._filesystems[host]

    @property
    def exists(self):
        status, response = self._fs.stat(self._url.path_with_params)
        if status.ok:
            return True
        elif status.errno == 3011:
            return False
        else:
            raise ValueError('Unrecognised error status: '+repr(status))

    def checksum(self, checksum_type=None):
        status, checksum = self._fs.query(client.flags.QueryCode.CHECKSUM, self._url.path_with_params)
        if status.ok:
            reponse_checksum_type, checksum = checksum.strip('\x00').split(' ')
            if checksum_type is not None and reponse_checksum_type != checksum_type:
                raise NotImplementedError(reponse_checksum_type, checksum_type)
            return checksum_type, int(checksum, base=16)
        elif status.errno == 3011:
            raise IOError(status.message)
        else:
            raise ValueError('Unrecognised error status: '+repr(status))

    def validate_checksum(self, other):
        if not self.exists:
            raise IOError(str(self) + "doesn't exist")
        if not other.exists:
            raise IOError(str(other) + "doesn't exist")
        checksum_type, self_checksum = self.checksum()
        _, other_checksum = other.checksum(checksum_type)
        if self_checksum != other_checksum:
            raise ChecksumMismatchError(self, self_checksum, other, other_checksum)

    def join(self, *args):
        return self.__class__(join(self._path, *args))


class CopyProgressHandler(client.utils.CopyProgressHandler):
    def __init__(self, total_jobs):
        self._jobs_pbar = tqdm(total=total_jobs)
        self._current_id = None
        self._current_pbar = None

    def begin(self, job_id, total, source, target):
        pass

    def end(self, job_id, results):
        self._current_id = None
        if self._current_pbar is not None:
            self._current_pbar.close()
            self._current_pbar = None
        self._jobs_pbar.update()

    def update(self, job_id, processed, total):
        try:
            if self._current_id is None:
                self._current_id = job_id
            elif self._current_id != job_id:
                print('Progress bar logic failure')
            elif self._current_pbar is None:
                # Create this the second loop to workaround for a race in tqdm
                self._current_pbar = tqdm(total=total, unit='B', unit_scale=True, leave=False)
            else:
                self._current_pbar.update(processed - self._current_pbar.n)
        except Exception as e:
            print('Exception in update', e)

    def should_cancel(self, job_id):
        print('Called CopyProgressHandler.should_cancel', job_id)
        return False


class CopyProcess(object):
    def __init__(self):
        self._jobs = {}

    def __bool__(self):
        return bool(self._jobs)

    def __len__(self):
        return len(self._jobs)

    def __contains__(self, lfn):
        return lfn in self._jobs

    def add_job(self, replica, destination):
        from .structures import Replica
        assert isinstance(replica, Replica)
        self._jobs[replica] = destination

    def copy(self):
        print('Copying', len(self._jobs), 'files')
        replicas = sorted(self._jobs, key=lambda replica: replica.lfn)

        process = client.CopyProcess()
        for replica in replicas:
            process.add_job(str(replica.pfn), str(self._jobs[replica]))

        status = process.prepare()
        assert status.ok, status

        status, results = process.run(CopyProgressHandler(len(self._jobs)))
        assert status.ok, (status, results)
        return {replica: result['status'].ok for replica, result in zip(replicas, results)}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__all__ = [
    'Replica',
]


class Replica(object):
    def __init__(self, lfn, se, pfn=None, error=None, banned=False):
        self._lfn = lfn
        self._se = se
        self._pfn = pfn
        self._error = error
        self._banned = banned

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._lfn == other._lfn and self._se == other._se
        return False

    def __str__(self):
        if self.available:
            return '{se}: {pfn}'.format(se=self._se, pfn=self._pfn)
        elif self._banned:
            return '{se}: Replica is banned'.format(se=self._se)
        elif self._error:
            return '{se}: {error}'.format(se=self._se, error=self._error)
        else:
            return '{se}: Unknown error ({error})'.format(se=self._se, error=repr(self))

    @property
    def available(self):
        return not self._banned and self._pfn is not None

    @property
    def lfn(self):
        return self._lfn

    @property
    def pfn(self):
        from .XRootD import URL
        return URL(self._pfn)

    @property
    def storage_element(self):
        return self._se



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import utils
# DIRAC must be initialised before importing other modules
utils.initalise_dirac()  # NOQA
from . import data_management
from . import exceptions
from . import XRootD
from .version import __version__


__all__ = [
    'data_management',
    'exceptions',
    'XRootD',
    '__version__',
]



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

from setuptools import find_packages
from setuptools import setup


def get_version():
    g = {}
    exec(open(os.path.join('manci', 'version.py')).read(), g)
    return g['__version__']


setup(name='manci',
      version=get_version(),
      packages=find_packages(exclude=['tests']),
      scripts=[],
      data_files=[],
      description='Convert between different style of formulae',
      long_description=None,
      author='Chris Burr',
      author_email='c.b@cern.ch',
      maintainer='Chris Burr',
      maintainer_email='c.b@cern.ch',
      url='https://github.com/chrisburr/manci',
      download_url='https://github.com/chrisburr/manci/releases',
      license='BSD 3-clause',
      test_suite='tests',
      install_requires=['numpy', 'pyparsing', 'colorlog', 'aenum'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Utilities',
          ],
      platforms='Any',
      )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__all__ = [
    'ChecksumMismatchError',
]


class ChecksumMismatchError(Exception):
    pass



# Licensed under a 3-clause BSD style license, see LICENSE.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__all__ = [
    '__version__',
    'version',
    'version_info',
]


__version__ = '0.0.0'
version = __version__
version_info = __version__.split('.')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import defaultdict
from os.path import basename
import sys

from .structures import Replica
from .utils import grouper


__all__ = [
    'lookup_files_from_query',
    'loookup_replicas',
    'mirror_files',
]


def lookup_files_from_query(bk_path):
    from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bkQuery = BKQuery(bkQuery=bk_path, visible='Yes')
    bkClient = BookkeepingClient()

    files = {}
    useFilesWithMetadata = False
    if useFilesWithMetadata:
        res = bkClient.getFilesWithMetadata(bkQuery.getQueryDict())
        if not res['OK']:
            print('ERROR getting the files', res['Message'], file=sys.stderr)
            sys.exit(1)
        parameters = res['Value']['ParameterNames']
        for record in res['Value']['Records']:
            dd = dict(zip(parameters, record))
            lfn = dd.pop('FileName')
            files[lfn] = dd

    else:
        lfns = bkQuery.getLFNs(printSEUsage=False, printOutput=False)

        for lfnChunk in grouper(lfns, 1000):
            res = bkClient.getFileMetadata(lfnChunk)
            if not res['OK']:
                print('ERROR: failed to get metadata:', res['Message'], file=sys.stderr)
                sys.exit(1)
            files.update(res['Value']['Successful'])

    if not files:
        print('No files found for BK query')
        sys.exit(0)

    print(len(files), 'files found')

    return files


def loookup_replicas(files, protocol=['xroot', 'root']):
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC.Resources.Storage.StorageElement import StorageElement
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    dm = DataManager()
    bk = BookkeepingClient()

    res = dm.getReplicas(list(files), getUrl=False)
    replicas = res.get('Value', {}).get('Successful', {})
    seList = sorted(set(se for lfn in files for se in replicas.get(lfn, {})))
    banned_SE_list = [se for se in seList if 'CNAF' in se]
    print('Found SE list of', seList)

    for lfn in files:
        files[lfn]['Replicas'] = []

    # Check if files are MDF
    bkRes = bk.getFileTypeVersion(list(files))
    assert not set(lfn for lfn, fileType in bkRes.get('Value', {}).iteritems() if fileType == 'MDF')
    for se in seList:
        # TODO Check if SEs are available
        lfns = [lfn for lfn in files if se in replicas.get(lfn, [])]

        if se in banned_SE_list:
            print('Skipping banned SE', se)
            for lfn in lfns:
                files[lfn]['Replicas'].append(Replica(lfn, se, banned=True))
            continue
        else:
            print('Looking up replicas for', len(lfns), 'files at', se)

        if lfns:
            res = StorageElement(se).getURL(lfns, protocol=protocol)
            if res['OK']:
                for lfn, pfn in res['Value']['Successful'].items():
                    files[lfn]['Replicas'].append(Replica(lfn, se, pfn=pfn))
                for lfn in res['Value']['Failed']:
                    files[lfn]['Replicas'].append(Replica(lfn, se, error=res))
            else:
                print('LFN -> PFN lookup failed for', se, 'with error:', res['Message'])
                for lfn in lfns:
                    files[lfn]['Replicas'].append(Replica(lfn, se, error=res['Message']))


def mirror_files(files, destination_dir, max_retries=1):
    from urban_barnacle import XRootD

    assert destination_dir.startswith('root://'), destination_dir
    assert len(files) == len(set(basename(lfn) for lfn in files)), 'Duplicate filenames found in input LFNs'

    n_tries = defaultdict(int)
    destination_dir = XRootD.URL(destination_dir)

    # Validate checksums of existing files
    for lfn, metadata in files.items():
        destination = destination_dir.join(basename(lfn))
        if destination.exists:
            validated = False
            for replica in metadata['Replicas']:
                if replica.available:
                    destination.validate_checksum(replica.pfn)
                    validated = True
                    break
            if not validated:
                print('Failed to validate checksum for', lfn, 'as no replicas are available')

    # Copy files which don't exist, repeating as needed
    keep_going = True
    while keep_going:
        copy_process = XRootD.CopyProcess()
        for lfn, metadata in files.items():
            destination = destination_dir.join(basename(lfn))
            if not destination.exists:
                for replica in metadata['Replicas']:
                    if not replica.available:
                        continue
                    if n_tries[replica] <= max_retries:
                        copy_process.add_job(replica, destination)
                        n_tries[replica] += 1
                        break

        copy_result = copy_process.copy()
        for replica, success in copy_result.items():
            destination = destination_dir.join(basename(replica.lfn))
            if success:
                destination.validate_checksum(replica.pfn)
            else:
                assert not destination.exists, (replica, destination)
        keep_going = copy_process and not all(copy_result.values())

    # Print the results
    n_successful = 0
    n_failed = 0
    for lfn, metadata in files.items():
        destination = destination_dir.join(basename(lfn))
        if destination.exists:
            n_successful += 1
        else:
            n_failed += 1
            print('Failed to copy', lfn, 'replicas are:')
            for replica in metadata['Replicas']:
                print(' >', replica)
            print()
    print(n_successful, 'out of', len(files), 'files copied successfully')
